from peptacular.fragment import fragment, sequence_trimmer, _trim_from_end, _trim_from_start, \
    build_fragment_sequences, create_fragment_internal_sequences

import unittest


class TestFragment(unittest.TestCase):
    def test_calculate_fragment_mz_values_with_unmodified_peptide(self):
        pyteomics_y1_frags = [346.1608765557, 249.10811270685, 120.06551961887999]
        pyteomics_y2_frags = [173.584076511235, 125.05769458680999, 60.536398042825]
        pyteomics_b1_frags = [98.06004031562, 227.10263340359, 328.150311872]
        pyteomics_b2_frags = [49.533658391195004, 114.05495493517999, 164.578794169385]
        y1_frags = list(fragment(sequence='PET', ion_types='y', charges=1))
        y2_frags = list(fragment(sequence='PET', ion_types='y', charges=2))
        b1_frags = list(fragment(sequence='PET', ion_types='b', charges=1))
        b2_frags = list(fragment(sequence='PET', ion_types='b', charges=2))

        self.assertEqual(len(pyteomics_y1_frags), len(y1_frags))
        for f in pyteomics_y1_frags:
            self.assertTrue(f in y1_frags)

        self.assertEqual(len(pyteomics_y2_frags), len(y2_frags))
        for f in pyteomics_y2_frags:
            self.assertTrue(f in y2_frags)

        self.assertEqual(len(pyteomics_b1_frags), len(b1_frags))
        for f in pyteomics_b1_frags:
            self.assertTrue(f in b1_frags)

        self.assertEqual(len(pyteomics_b2_frags), len(b2_frags))
        for f in pyteomics_b2_frags:
            self.assertTrue(f in b2_frags)

    def test_calculate_fragment_mz_values_with_modified_peptide(self):
        pyteomics_y_frags = [174.584076511235, 125.05769458680999, 60.536398042825]
        pyteomics_b_frags = [50.533658391195004, 115.05495493517999, 165.578794169385]

        y_frags = list(fragment(sequence='P(2)ET', ion_types='y', charges=2))
        b_frags = list(fragment(sequence='P(2)ET', ion_types='b', charges=2))

        self.assertEqual(len(pyteomics_y_frags), len(y_frags))
        for f in pyteomics_y_frags:
            self.assertTrue(f in y_frags)

        self.assertEqual(len(pyteomics_b_frags), len(b_frags))
        for f in pyteomics_b_frags:
            self.assertTrue(f in b_frags)

    def test_calculate_fragment_mz_values_with_all_ion_types(self):
        pyteomics = {'a': [70.06512569606001, 199.10771878403003, 300.15539725243997],
                     'b': [98.06004031562, 227.10263340359, 328.150311872],
                     'c': [115.08658941662999, 244.1291825046, 345.17686097301],
                     'x': [372.14014111112, 275.08737726226997, 146.0447841743],
                     'y': [346.1608765557, 249.10811270685, 120.06551961887999],
                     'z': [329.13432745469, 232.08156360584, 103.03897051786998],
                     }

        for ion_type in 'abcxyz':
            frags = sorted(list(fragment(sequence='PET', ion_types=ion_type, charges=1)))
            pyteomics_frags = sorted(pyteomics[ion_type])
            for f, pf in zip(frags, pyteomics_frags):
                self.assertAlmostEqual(f, pf, 6)

    def test_sequence_trimmer(self):
        # Test forward generator
        generator = sequence_trimmer("A(P)C(P)G", True)
        self.assertEqual(next(generator), "A(P)C(P)G")
        self.assertEqual(next(generator), "C(P)G")
        self.assertEqual(next(generator), "G")
        with self.assertRaises(StopIteration):
            next(generator)

        # Test backward generator
        generator = sequence_trimmer("A(P)C(P)G", False)
        self.assertEqual(next(generator), "A(P)C(P)G")
        self.assertEqual(next(generator), "A(P)C(P)")
        self.assertEqual(next(generator), "A(P)")
        with self.assertRaises(StopIteration):
            next(generator)

    def test_trim_from_start(self):
        generator = _trim_from_end("A(P)C(P)G")
        self.assertEqual(next(generator), "A(P)C(P)G")
        self.assertEqual(next(generator), "A(P)C(P)")
        self.assertEqual(next(generator), "A(P)")
        with self.assertRaises(StopIteration):
            next(generator)

    def test_trim_from_end(self):
        generator = _trim_from_start("[P]A(P)C(P)G")
        self.assertEqual(next(generator), "[P]A(P)C(P)G")
        self.assertEqual(next(generator), "C(P)G")
        self.assertEqual(next(generator), "G")
        with self.assertRaises(StopIteration):
            next(generator)

    def test_create_fragment_sequences(self):
        sequence = "PEPTIDE"

        for ion_type in 'abc':
            fragments = build_fragment_sequences(sequence, ion_type)
            self.assertEqual(fragments, ['P', 'PE', 'PEP', 'PEPT', 'PEPTI', 'PEPTID', 'PEPTIDE'][::-1])

        for ion_type in 'xyz':
            fragments = build_fragment_sequences(sequence, ion_type)
            self.assertEqual(fragments, ['E', 'DE', 'IDE', 'TIDE', 'PTIDE', 'EPTIDE', 'PEPTIDE'][::-1])

        modified_sequence = "[-10]PEP(2)TIDE(100)"

        for ion_type in 'abc':
            fragments = build_fragment_sequences(modified_sequence, ion_type)
            self.assertEqual(fragments,
                             ['[-10]P', '[-10]PE', '[-10]PEP(2)', '[-10]PEP(2)T', '[-10]PEP(2)TI', '[-10]PEP(2)TID',
                              '[-10]PEP(2)TIDE(100)'][::-1])

        for ion_type in 'xyz':
            fragments = build_fragment_sequences(modified_sequence, ion_type)
            self.assertEqual(fragments,
                             ['E(100)', 'DE(100)', 'IDE(100)', 'TIDE(100)', 'P(2)TIDE(100)', 'EP(2)TIDE(100)',
                              '[-10]PEP(2)TIDE(100)'][::-1])

    def test_create_internal_sequences(self):
        sequence = "PEPTIDE"

        for ion_type in 'xyz':
            fragments = create_fragment_internal_sequences(sequence, ion_type)
            self.assertEqual(fragments, ['P', 'PE', 'PEP', 'PEPT', 'PEPTI', 'PEPTID', 'PEPTIDE'][::-1])

        for ion_type in 'abc':
            fragments = create_fragment_internal_sequences(sequence, ion_type)
            self.assertEqual(fragments, ['E', 'DE', 'IDE', 'TIDE', 'PTIDE', 'EPTIDE', 'PEPTIDE'][::-1])

        modified_sequence = "[-10]PEP(2)TIDE(100)"

        for ion_type in 'xyz':
            fragments = create_fragment_internal_sequences(modified_sequence, ion_type)
            self.assertEqual(fragments,
                             ['[-10]P', '[-10]PE', '[-10]PEP(2)', '[-10]PEP(2)T', '[-10]PEP(2)TI', '[-10]PEP(2)TID',
                              '[-10]PEP(2)TIDE(100)'][::-1])

        for ion_type in 'abc':
            fragments = create_fragment_internal_sequences(modified_sequence, ion_type)
            self.assertEqual(fragments,
                             ['E(100)', 'DE(100)', 'IDE(100)', 'TIDE(100)', 'P(2)TIDE(100)', 'EP(2)TIDE(100)',
                              '[-10]PEP(2)TIDE(100)'][::-1])


if __name__ == "__main__":
    unittest.main()



from peptacular.sequence import get_modifications, add_modifications, strip_modifications, \
    apply_static_modifications, apply_variable_modifications

import unittest


class TestSequence(unittest.TestCase):

    def test_read_modified_peptide(self):
        # Test the case where there are no modifications in the peptide sequence
        peptide_sequence = "ACDEFG"
        expected_output = {}
        self.assertEqual(get_modifications(peptide_sequence), expected_output)

        # Test the case where there is a single modification in the peptide sequence
        peptide_sequence = "A(2)CDEFG"
        expected_output = {0: 2}
        self.assertEqual(get_modifications(peptide_sequence), expected_output)

        # Test the case where there are multiple modifications in the peptide sequence
        peptide_sequence = "A(2)C(3)DE(1)FG"
        expected_output = {0: 2, 1: 3, 3: 1}
        self.assertEqual(get_modifications(peptide_sequence), expected_output)

    def test_create_modified_peptide(self):
        # Test the case where there are no modifications in the peptide sequence
        unmodified_sequence = "ACDEFG"
        modifications = {}
        expected_output = "ACDEFG"
        self.assertEqual(add_modifications(unmodified_sequence, modifications), expected_output)

        # Test the case where there is a single modification in the peptide sequence
        unmodified_sequence = "ACDEFG"
        modifications = {0: 2}
        expected_output = "A(2)CDEFG"
        self.assertEqual(add_modifications(unmodified_sequence, modifications), expected_output)

        # Test the case where there is a single N-Term mod
        unmodified_sequence = "ACDEFG"
        modifications = {-1: 2}
        expected_output = "[2]ACDEFG"
        self.assertEqual(add_modifications(unmodified_sequence, modifications), expected_output)

        # Test the case where there is a single C-Term mod
        unmodified_sequence = "ACDEFG"
        modifications = {5: 2}
        expected_output = "ACDEFG(2)"
        self.assertEqual(add_modifications(unmodified_sequence, modifications), expected_output)

        # Test the case where there are multiple modifications in the peptide sequence
        unmodified_sequence = "ACDEFG"
        modifications = {0: 2, 1: 3, 3: 1}
        expected_output = "A(2)C(3)DE(1)FG"
        self.assertEqual(add_modifications(unmodified_sequence, modifications), expected_output)

        # Test the case where the index of a modification is invalid for the given peptide sequence
        unmodified_sequence = "ACDEFG"
        modifications = {10: 2}
        with self.assertRaises(ValueError):
            add_modifications(unmodified_sequence, modifications)

    def test_get_unmodified_peptide(self):
        self.assertEqual(strip_modifications('ACDEFG'), 'ACDEFG')
        self.assertEqual(strip_modifications('A(2)CDE(2)FG'), 'ACDEFG')
        self.assertEqual(strip_modifications('A(2)CDE(Acetyl)FG'), 'ACDEFG')

    def test_add_static_mod(self):
        self.assertEqual(apply_static_modifications('PEPCTIDE', {'C': 57.021464}), 'PEPC(57.021464)TIDE')
        self.assertEqual(apply_static_modifications('PEPC(57.021464)TIDE', {'C': 57.021464}), 'PEPC(57.021464)TIDE')
        self.assertEqual(apply_static_modifications('CPEPTIDEC', {'C': 57.021464}), 'C(57.021464)PEPTIDEC(57.021464)')
        self.assertEqual(apply_static_modifications('P(1)EPCTIDE(1)', {'C': 57.021464}), 'P(1)EPC(57.021464)TIDE(1)')

    def test_add_variable_mod(self):
        self.assertEqual(set(apply_variable_modifications('PEPCTIDCE', {'C': 57.021464}, 2)), {'PEPC(57.021464)TIDCE',
                                                                                    'PEPC(57.021464)TIDC(57.021464)E',
                                                                                    'PEPCTIDC(57.021464)E',
                                                                                    'PEPCTIDCE'})

        self.assertEqual(set(apply_variable_modifications('P(1)EPC(20)TIDCE', {'C': 57.021464}, 2)), {'P(1)EPC(20)TIDCE',
                                                                                           'P(1)EPC(20)TIDC(57.021464)E'})


if __name__ == "__main__":
    unittest.main()



"""
Used in digest.py, and exclusively work with unmodified sequences!
"""

from typing import Tuple, List
from itertools import groupby


def build_non_enzymatic_spans(span: Tuple[int, int, int], min_len: int = 1, max_len: int = None) \
        -> List[Tuple[int, int, int]]:
    """
    Generates and returns all possible sub-spans of the given span.

    :param span: A tuple representing the original span.
    :type span: Tuple[int, int, int]
    :param min_len: The minimum length of sub-spans to be generated.
    :type min_len: int
    :param max_len: The maximum length of sub-spans to be generated.
    :type max_len: int

    :return: A list of all possible sub-spans as tuples.
    :rtype: List[Tuple[int, int, int]]

    .. code-block:: python

        >>> build_non_enzymatic_spans((0, 3, 0), 1, 2)
        [(0, 1, 0), (0, 2, 0), (1, 2, 0), (1, 3, 0), (2, 3, 0)]

        >>> build_non_enzymatic_spans((0, 3, 0))
        [(0, 1, 0), (0, 2, 0), (1, 2, 0), (1, 3, 0), (2, 3, 0)]

        >>> build_non_enzymatic_spans((0, 3, 0), 1, 10)
        [(0, 1, 0), (0, 2, 0), (1, 2, 0), (1, 3, 0), (2, 3, 0)]

    """

    if min_len is None:
        min_len = 1

    max_span = span[1] - span[0] - 1
    if max_len is None:
        max_len = max_span
    max_len = min(max_len, max_span)

    start, end, _ = span
    return [(i, j, 0) for i in range(start, end + 1) for j in range(i + min_len, min(end + 1, i + max_len + 1))]


def build_left_semi_spans(span: Tuple[int, int, int], min_len: int = 1, max_len: int = None) \
        -> List[Tuple[int, int, int]]:
    """
    Generates and returns all possible sub-spans of the given span starting from the left.

    :param span: A tuple representing the original span.
    :type span: Tuple[int, int, int]
    :param min_len: The minimum length of sub-spans to be generated.
    :type min_len: int
    :param max_len: The maximum length of sub-spans to be generated.
    :type max_len: int

    :return: A list of all possible left sub-spans as tuples.
    :rtype: List[Tuple[int, int, int]]

    .. code-block:: python

        >>> build_left_semi_spans((0, 3, 0))
        [(0, 2, 0), (0, 1, 0)]

        # Keeps the value of the original span
        >>> build_left_semi_spans((0, 3, 2))
        [(0, 2, 2), (0, 1, 2)]

        # Set min and max length
        >>> build_left_semi_spans((0, 3, 0), 1, 1)
        [(0, 1, 0)]
        >>> build_left_semi_spans((0, 3, 0), 1, 10)
        [(0, 2, 0), (0, 1, 0)]

    """
    if min_len is None:
        min_len = 1

    if max_len is None:
        max_len = span[1] - span[0]

    start, end, value = span
    new_end = min(start + max_len, end - 1)
    return [(start, i, value) for i in range(new_end, start - 1, -1) if i - start >= min_len]


def build_right_semi_spans(span: Tuple[int, int, int], min_len: int = 1, max_len: int = None) \
        -> List[Tuple[int, int, int]]:
    """
    Generates and returns all possible sub-spans of the given span starting from the right.

    :param span: A tuple representing the original span.
    :type span: Tuple[int, int, int]
    :param min_len: The minimum length of sub-spans to be generated.
    :type min_len: int
    :param max_len: The maximum length of sub-spans to be generated.
    :type max_len: int

    :return: A list of all possible right sub-spans as tuples.
    :rtype: List[Tuple[int, int, int]]

    .. code-block:: python

        >>> build_right_semi_spans((0, 3, 0))
        [(1, 3, 0), (2, 3, 0)]

        # Keeps the value of the original span
        >>> build_right_semi_spans((0, 3, 2))
        [(1, 3, 2), (2, 3, 2)]

        # Set min and max length
        >>> build_right_semi_spans((0, 3, 0), 1, 1)
        [(2, 3, 0)]
        >>> build_right_semi_spans((0, 3, 0), 1, 10)
        [(1, 3, 0), (2, 3, 0)]

    """

    if min_len is None:
        min_len = 1

    if max_len is None:
        max_len = span[1] - span[0]

    start, end, value = span
    new_start = max(start + 1, end - max_len)
    return [(i, end, value) for i in range(new_start, end + 1) if end - i >= min_len]


def build_enzymatic_spans(max_index: int, enzyme_sites: List[int], missed_cleavages: int,
                          min_len: int = 1, max_len: int = None) -> List[Tuple[int, int, int]]:
    """
    Computes enzymatic spans for the given enzyme sites and missed cleavages.

    :param max_index: The max index of the span.
    :type max_index: int
    :param enzyme_sites: The list of indices representing enzyme sites.
    :type enzyme_sites: List[int]
    :param missed_cleavages: The number of allowed missed cleavages.
    :type missed_cleavages: int
    :param min_len: The minimum length of an enzymatic span.
    :type min_len: int, optional
    :param max_len: The maximum length of an enzymatic span.
    :type max_len: int, optional

    :return: A list of tuples representing enzymatic spans.
    :rtype: List[Tuple[int, int, int]]

    .. code-block:: python

        >>> build_enzymatic_spans(5, [0, 3, 5], 1)
        [(0, 3, 0), (0, 5, 1), (3, 5, 0)]

        # Set min length
        >>> build_enzymatic_spans(5, [0, 3, 5], 1, min_len=5)
        [(0, 5, 1)]

        # Set max length
        >>> build_enzymatic_spans(5, [0, 3, 5], 1, max_len=3)
        [(0, 3, 0), (3, 5, 0)]

    """
    if min_len is None:
        min_len = 1

    if max_len is None:
        max_len = max_index

    enzyme_sites = set(enzyme_sites)
    if 0 not in enzyme_sites:
        enzyme_sites.add(0)

    if max_index not in enzyme_sites:
        enzyme_sites.add(max_index)

    enzyme_sites = sorted(list(enzyme_sites))

    spans = []
    for i, start_site in enumerate(enzyme_sites):
        spans.extend(
            [(start_site, end_site, j) for j, end_site in enumerate(enzyme_sites[i + 1:i + missed_cleavages + 2])])

    # Filter spans based on length
    spans = [span for span in spans if min_len <= span[1] - span[0] <= max_len]

    return spans


def _grouped_left_semi_span_builder(spans: List[Tuple[int, int, int]], min_len: int, max_len: int) -> \
        List[Tuple[int, int, int]]:
    """
    Get all left semi-spans from the given list of spans that have a length within the specified range.

    This function groups the spans by the start position, for each group, it checks every span to see if its length
    is at least the minimum length. If it is, it calculates the new maximum length, which is the smallest of the max_len
    and the span's length. Then it adds the left semi-spans to the list of semi-spans.

    :param spans: A list of tuples, where each tuple represents a span with three integers (start, end, value).
    :type spans: List[Tuple[int, int, int]]
    :param min_len: The minimum length of the left semi-spans.
    :type min_len: int
    :param max_len: The maximum length of the left semi-spans.
    :type max_len: int

    :return: A list of tuples representing all left semi-spans that are within the specified length range.
    :rtype: List[Tuple[int, int, int]]

    .. code-block:: python

        >>> _grouped_left_semi_span_builder([(0, 3, 0), (0, 5, 1), (3, 5, 0)], 1, 5)
        [(0, 4, 1), (0, 2, 0), (0, 1, 0), (3, 4, 0)]

    """
    semi_spans = []
    spans = sorted(spans, key=lambda x: (x[0], -x[2]))
    for _, group in groupby(spans, key=lambda x: x[0]):
        group = list(group)
        for i, span in enumerate(group):
            span_len = span[1] - span[0]

            if span_len <= min_len:
                break

            new_max_len = min(max_len, span_len - 1)
            if i == len(group) - 1:
                semi_spans.extend(build_left_semi_spans(span, min_len, new_max_len))
            else:
                next_span = group[i + 1]
                next_span_len = next_span[1] - next_span[0]
                new_min = max(min_len, next_span_len + 1)
                semi_spans.extend(build_left_semi_spans(span, new_min, new_max_len))

    return semi_spans


def _grouped_right_semi_span_builder(spans: List[Tuple[int, int, int]], min_len: int, max_len: int) -> \
        List[Tuple[int, int, int]]:
    """
    Get all right semi-spans from the given list of spans that have a length within the specified range.

    This function groups the spans by the start position, for each group, it checks every span to see if its length
    is at least the minimum length. If it is, it calculates the new maximum length, which is the smallest of the max_len
    and the span's length. Then it adds the right semi-spans to the list of semi-spans.

    :param spans: A list of tuples, where each tuple represents a span with three integers (start, end, value).
    :type spans: List[Tuple[int, int, int]]
    :param min_len: The minimum length of the right semi-spans.
    :type min_len: int
    :param max_len: The maximum length of the right semi-spans.
    :type max_len: int

    :return: A list of tuples representing all right semi-spans that are within the specified length range.
    :rtype: List[Tuple[int, int, int]]

    .. code-block:: python

        >>> _grouped_right_semi_span_builder([(0, 3, 0), (0, 5, 1), (3, 5, 0)], 1, 5)
        [(1, 3, 0), (2, 3, 0), (1, 5, 1), (2, 5, 1), (4, 5, 0)]

    """

    semi_spans = []
    spans = sorted(spans, key=lambda x: (x[1], -x[2]))
    for _, group in groupby(spans, key=lambda x: x[1]):
        group = list(group)
        for i, span in enumerate(group):
            span_len = span[1] - span[0]

            if span_len < min_len:
                break

            new_max_len = min(max_len, span_len - 1)
            if i == len(group) - 1:
                semi_spans.extend(build_right_semi_spans(span, min_len, new_max_len))
            else:
                next_span = group[i + 1]
                next_span_len = next_span[1] - next_span[0]
                new_min = max(min_len, next_span_len + 1)
                semi_spans.extend(build_right_semi_spans(span, new_min, new_max_len))

    return semi_spans


def build_semi_spans(spans: List[Tuple[int, int, int]], min_len: int = 1, max_len: int = None) -> \
        List[Tuple[int, int, int]]:
    """
    Computes semi spans for a given list of spans based on length criteria.

    :param spans: The list of spans.
    :type spans: List[Tuple[int, int, int]]
    :param min_len: The minimum length of a semi span.
    :type min_len: int
    :param max_len: The maximum length of a semi span.
    :type max_len: int

    :return: A list of tuples representing semi spans.
    :rtype: List[Tuple[int, int, int]]

    .. code-block:: python

        >>> build_semi_spans([(0, 3, 0), (0, 5, 1), (3, 5, 0)], 1, 5)
        [(0, 4, 1), (0, 2, 0), (0, 1, 0), (3, 4, 0), (1, 3, 0), (2, 3, 0), (1, 5, 1), (2, 5, 1), (4, 5, 0)]

    """

    if min_len is None:
        min_len = 1

    if max_len is None:
        max_len = max(span[1] - span[0] for span in spans)

    return _grouped_left_semi_span_builder(spans, min_len, max_len) + _grouped_right_semi_span_builder(spans, min_len, max_len)


def build_spans(max_index: int, enzyme_sites: List[int], missed_cleavages: int, min_len: int,
                max_len: int, semi: bool) -> List[Tuple[int, int, int]]:
    """
    Builds spans for a given sequence based on enzyme sites, missed cleavages, and length criteria.

    :param max_index: The max index of the span.
    :type max_index: int
    :param enzyme_sites: The list of indices representing enzyme sites.
    :type enzyme_sites: List[int]
    :param missed_cleavages: The number of allowed missed cleavages.
    :type missed_cleavages: int
    :param min_len: The minimum length of a span.
    :type min_len: int
    :param max_len: The maximum length of a span.
    :type max_len: int
    :param semi: Whether to compute and include semi spans.
    :type semi: bool

    :return: A list of tuples representing spans.
    :rtype: List[Tuple[int, int, int]]
    """

    if min_len is None:
        min_len = 1

    if max_len is None:
        max_len = max_index

    if len(enzyme_sites) == max_index:  # non-enzymatic
        return build_non_enzymatic_spans((0, max_index, 0), min_len, max_len)

    if semi:
        spans = build_enzymatic_spans(max_index, enzyme_sites, missed_cleavages, min_len, None)
        semi_spans = build_semi_spans(spans, min_len, max_len)

        # filter spans by len
        spans = [span for span in spans if max_len >= span[1] - span[0] >= min_len] + semi_spans
    else:
        spans = build_enzymatic_spans(max_index, enzyme_sites, missed_cleavages, min_len, max_len)
    return spans


def calculate_span_coverage(spans: List[Tuple[int, int, int]], max_index: int, accumulate: bool = False) -> List[int]:
    """
    Calculates the coverage array for a given list of spans.

    :param spans: The list of spans.
    :type spans: List[Tuple[int, int, int]]
    :param max_index: The max index of the span.
    :type max_index: int
    :param accumulate: Whether to accumulate the coverage array.
    :type accumulate: bool

    :return: The coverage array.
    :rtype: List[int]

    .. code-block:: python

        >>> calculate_span_coverage([(0, 3, 0), (3, 6, 0), (6, 9, 0)], 9)
        [1, 1, 1, 1, 1, 1, 1, 1, 1]

        >>> calculate_span_coverage([(0, 3, 0), (0, 3, 0), (6, 9, 0)], 9)
        [1, 1, 1, 0, 0, 0, 1, 1, 1]

        >>> calculate_span_coverage([(0, 3, 0), (0, 3, 0), (6, 9, 0)], 9, accumulate=True)
        [2, 2, 2, 0, 0, 0, 1, 1, 1]

        >>> calculate_span_coverage([(0, 3, 0), (3, 6, 0), (6, 9, 0)], 12)
        [1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0]

        >>> calculate_span_coverage([(0, 3, 0), (3, 6, 0), (6, 9, 0)], 6)
        Traceback (most recent call last):
        ...
        IndexError: list assignment index out of range

    """

    cov_array = [0] * max_index
    for span in spans:
        for i in range(span[0], span[1]):
            if accumulate:
                cov_array[i] += 1
            else:
                cov_array[i] = 1

    return cov_array




from peptacular.constants import MONO_ISOTOPIC_ATOMIC_MASSES, AVERAGE_ATOMIC_MASSES, AVERAGE_AA_MASSES, \
    MONO_ISOTOPIC_AA_MASSES, ION_ADJUSTMENTS, UWPR_MONO_ISOTOPIC_ATOMIC_MASSES, UWPR_AVERAGE_AA_MASSES, \
    UWPR_AVERAGE_ATOMIC_MASSES, UWPR_MONO_ISOTOPIC_AA_MASSES
from peptacular.sequence import get_modifications, strip_modifications, is_sequence_valid
from peptacular.util import validate_ion_type


def calculate_mass(sequence: str, charge: int = 0, ion_type: str = 'y', monoisotopic: bool = True,
                   uwpr_mass: bool = False) -> float:
    """
    Calculate the mass of an amino acid 'sequence'.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param charge: The charge state, default is [0].
    :type charge: int
    :param ion_type: The ion type, default is [y].
    :type ion_type: str
    :param monoisotopic: If True, use monoisotopic mass else use average mass, defaults to [True].
    :type monoisotopic: bool
    :param uwpr_mass: If true, uses uwpr masses. If false, uses Pyteomics masses, defaults to [False].
    :type uwpr_mass: bool

    :raise ValueError: If the ion type is not one of 'a', 'b', 'c', 'x', 'y', or 'z'.

    :return: Mass of the peptide sequence.
    :rtype: float

    .. code-block:: python

        # Calculate the mass of a peptide sequence.
        >>> calculate_mass('PEPTIDE')
        799.3599640267099

        # Calculate the b-ion mass of a peptide sequence.
        >>> calculate_mass('PEPTIDE', ion_type='b')
        781.34939934301

        # Calulate the average mass of a peptide sequence.
        >>> calculate_mass('PEPTIDE', monoisotopic=False)
        799.8280646837

        # Calculate the mass of a peptide sequence with a charge of 2.
        >>> calculate_mass('PEPTIDE', charge=2)
        801.3745169602499

        # Calcualte the mass of a modified peptide sequence.
        >>> calculate_mass('PE(3.14)PTIDE[80]')
        882.4999640267099

    """

    validate_ion_type(ion_type=ion_type)

    # Select mass set based on monoisotopic flag
    atomic_masses, aa_masses = (MONO_ISOTOPIC_ATOMIC_MASSES, MONO_ISOTOPIC_AA_MASSES) \
        if monoisotopic is True else (AVERAGE_ATOMIC_MASSES, AVERAGE_AA_MASSES)

    if uwpr_mass is True:
        atomic_masses, aa_masses = (UWPR_MONO_ISOTOPIC_ATOMIC_MASSES, UWPR_MONO_ISOTOPIC_AA_MASSES) \
            if monoisotopic is True else (UWPR_AVERAGE_ATOMIC_MASSES, UWPR_AVERAGE_AA_MASSES)

    # Parse modifications and strip them from sequence
    mods = get_modifications(sequence=sequence)
    stripped_sequence = strip_modifications(sequence=sequence)

    # Calculate mass
    mass = sum(aa_masses[aa] for aa in stripped_sequence)
    mass += sum(float(value) for value in mods.values())
    mass += (charge * atomic_masses['PROTON'])

    return mass + ION_ADJUSTMENTS[ion_type]


def calculate_mz(sequence: str, charge: int = 0, ion_type: str = 'y',
                 monoisotopic: bool = True, uwpr_mass: bool = False) -> float:
    """
    Calculate the m/z (mass-to-charge ratio) of an amino acid 'sequence'.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param charge: The charge state, default is [0].
    :type charge: int
    :param ion_type: The ion type, default is [y].
    :type ion_type: str
    :param monoisotopic: If True, use monoisotopic mass else use average mass, defaults to [True].
    :type monoisotopic: bool
    :param uwpr_mass: If true, uses uwpr masses. If false, uses Pyteomics masses, defaults to [False].
    :type uwpr_mass: bool

    :raise ValueError: If the ion type is not one of 'a', 'b', 'c', 'x', 'y', or 'z'.

    :return: m/z of the peptide sequence.
    :rtype: float

    .. code-block:: python

        # Calculate the m/z of a peptide sequence.
        >>> calculate_mz('PEPTIDE', charge = 1)
        800.3672404934799

        # Calculate the b-ion m/z of a peptide sequence.
        >>> calculate_mz('PEPTIDE', charge = 1, ion_type='b')
        782.35667580978

        # Calulate the average m/z of a peptide sequence.
        >>> calculate_mz('PEPTIDE', charge = 1, monoisotopic=False)
        800.83534115047

        # Calculate the m/z of a peptide sequence with a charge of 2.
        >>> calculate_mz('PEPTIDE', charge=2)
        400.68725848012497

        # Calcualte the m/z of a modified peptide sequence.
        >>> calculate_mz('PE(3.14)PTIDE[80]', charge=2)
        442.25725848012496

    """

    mass = calculate_mass(sequence=sequence, charge=charge, ion_type=ion_type,
                          monoisotopic=monoisotopic, uwpr_mass=uwpr_mass)
    return mass if charge == 0 else mass / charge


def valid_mass_sequence(sequence: str):
    """
    Check if a sequence is a valid mass sequence.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str

    :return: True if the sequence is a valid mass sequence, False otherwise.
    :rtype: bool

    .. code-block:: python

        # Check if a sequence is a valid mass sequence.
        >>> valid_mass_sequence('PEPTIDE')
        True
        >>> valid_mass_sequence('PEPTIDE[80]')
        True
        >>> valid_mass_sequence('PEPTIDE[80.0]')
        True
        >>> valid_mass_sequence('PEPTIDE(Ox)PEPTIDE')
        False

    """

    if is_sequence_valid(sequence) is False:
        return False

    stripped_sequence = strip_modifications(sequence)
    mods = get_modifications(sequence)

    for aa in stripped_sequence:
        if aa not in MONO_ISOTOPIC_AA_MASSES:
            return False

    for _, mod in mods.items():
        if isinstance(mod, float) is False and isinstance(mod, int) is False:
            return False

    return True



from dataclasses import dataclass
from typing import List, Generator, Union

from peptacular.constants import PROTON_MASS
from peptacular.mass import calculate_mz
from peptacular.sequence import calculate_sequence_length
from peptacular.term import strip_c_term, strip_n_term
from peptacular.util import is_forward


@dataclass(frozen=True)
class Fragment:
    """
    A dataclass for representing a peptide fragment ion.

    :ivar str sequence: The fragment sequence, including any modifications.
    :ivar float mz: The fragment mass-to-charge (m/z) ratio
    :ivar int charge: The fragment charge state.
    :ivar str ion_type: The fragment ion type.
    :ivar int number: The fragment ion number, for abc ions this refers to the number of residues starting from the
                      N-terminus, for xyz ions this refers to the number of residues starting from the C-terminus.
    :ivar bool internal: True if the fragment is an internal fragment, False if it is a terminal fragment.
    :ivar int parent_number: The number of the parent fragment from which the internal fragment is derived. For
                             terminal fragments (internal=False) the number and parent number should always be equal.

    .. code-block:: python

        >>> frag = Fragment("PE(3.1415)PTIDE", 123.456, 1, "b", 2, False, 2)
        >>> frag.sequence
        'PE(3.1415)PTIDE'
        >>> frag.mz
        123.456
        >>> frag.charge
        1
        >>> frag.ion_type
        'b'
        >>> frag.number
        2
        >>> frag.internal
        False
        >>> frag.parent_number
        2

    """

    sequence: str
    mz: float
    charge: int
    ion_type: str
    number: int
    internal: bool
    parent_number: int

    def __post_init__(self):
        """
        Validates the fragment after initialization.
        """

        if self.internal is False and self.number != self.parent_number:
            raise ValueError("Parent number must be equal to number for terminal fragments.")

    @property
    def start(self):
        """
        Returns the start index of the fragment sequence in the parent sequence.

        :return: Start index of the fragment.
        :rtype: int

        :raises ValueError: If the ion type is invalid.
        """

        if self.ion_type in 'abc':
            if self.internal is False:
                return 0
            return self.parent_number - self.number

        if self.ion_type in 'xyz':
            return -self.parent_number

        raise ValueError(f"Invalid ion type: {self.ion_type}")

    @property
    def end(self):
        """
        Returns the end index of the fragment sequence in the parent sequence.

        :return: End index of the fragment.
        :rtype: int

        :raises ValueError: If the ion type is invalid.
        """

        if self.ion_type in 'abc':
            return self.parent_number

        if self.ion_type in 'xyz':
            if self.internal is False:
                return None
            return self.start + self.number

        raise ValueError(f"Invalid ion type: {self.ion_type}")

    @property
    def mass(self):
        """
        Returns the mass of the fragment.

        :return: Mass of the fragment.
        :rtype: float
        """

        return self.mz * self.charge

    @property
    def neutral_mass(self):
        """
        Returns the neutral mass of the fragment.

        :return: Neutral mass of the fragment.
        :rtype: float
        """

        return self.mass - self.charge * PROTON_MASS

    @property
    def label(self):
        """
        Returns the label of the fragment, e.g., b2, y3i, etc.

        :return: Label of the fragment.
        :rtype: str
        """

        return f"{'+' * self.charge}" \
               f"{self.ion_type}" \
               f"{self.parent_number}" \
               f"{'i' if self.internal else ''}" \
               f"{self.number if self.internal else ''}"


def build_fragments(sequence: str, ion_types: Union[List[str], str], charges: Union[List[int], int],
                    monoisotopic: bool = True, internal: bool = False) -> List[Fragment]:
    """
    Builds all Fragment objects or a given input 'sequence'.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param ion_types: A list of ion types to consider, e.g., ['b', 'y'], or a single ion type, e.g., 'b'.
    :type ion_types: Union[List[str], str]
    :param charges: A list of charge states for the fragments, or a single charge state.
    :type charges: Union[List[int], int]
    :param monoisotopic: If True, use monoisotopic masses. If False, use average masses, default is [True].
    :type monoisotopic: bool
    :param internal: If True, include internal fragments. If False, only include terminal fragments, default is [False].
    :type internal: bool

    :return: List of Fragment objects.
    :rtype: List[Fragment]

    .. code-block:: python

        # Create a list of y ions for a modified peptide.
        >>> frags = build_fragments("[1.0]P(2.0)E(3.0)[4.0]", 'y', 1, monoisotopic=True, internal=False)
        >>> len(frags)
        2

        >>> frags[0].sequence
        '[1.0]P(2.0)E(3.0)[4.0]'

        >>> frags[0].mz
        255.11319808729002

        >>> frags[1].sequence
        'E(3.0)[4.0]'

        >>> frags[1].mz
        155.06043423844

    """

    if isinstance(ion_types, str):
        ion_types = [ion_types]

    if isinstance(charges, int):
        charges = [charges]

    fragments = []
    for ion_type in ion_types:
        seq_length = calculate_sequence_length(sequence)
        for number, fragment_sequence in enumerate(build_fragment_sequences(sequence=sequence, ion_type=ion_type)):

            frag_number = seq_length - number
            for charge in charges:
                mz = calculate_mz(fragment_sequence, charge=charge, ion_type=ion_type, monoisotopic=monoisotopic)
                fragments.append(Fragment(sequence=fragment_sequence, mz=mz, charge=charge, ion_type=ion_type,
                                          number=frag_number, internal=False, parent_number=frag_number))
            if internal is True:
                fragment_seq_len = calculate_sequence_length(fragment_sequence)
                for internal_number, internal_fragment_sequence in enumerate(
                        create_fragment_internal_sequences(sequence=fragment_sequence, ion_type=ion_type)):

                    if internal_number == 0:  # skip first since its already created
                        continue

                    internal_frag_number = fragment_seq_len - internal_number

                    for charge in charges:
                        mz = calculate_mz(internal_fragment_sequence, charge=charge, ion_type=ion_type,
                                          monoisotopic=monoisotopic)
                        fragments.append(
                            Fragment(sequence=internal_fragment_sequence, mz=mz, charge=charge, ion_type=ion_type,
                                     number=internal_frag_number, internal=True, parent_number=frag_number))

    return fragments


def build_fragment_sequences(sequence: str, ion_type: str) -> List[str]:
    """
    Builds all fragment subsequences for the input `sequence`.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param ion_type: The type of ion for which fragments are generated.
    :type ion_type: str

    :raise ValueError: If the ion type is not one of 'a', 'b', 'c', 'x', 'y', or 'z'.

    :return: The fragment sequences.
    :rtype: List[str]

    .. code-block:: python

        >>> build_fragment_sequences("PEPTIDE", "a")
        ['PEPTIDE', 'PEPTID', 'PEPTI', 'PEPT', 'PEP', 'PE', 'P']

        >>> build_fragment_sequences("PEPTIDE", "y")
        ['PEPTIDE', 'EPTIDE', 'PTIDE', 'TIDE', 'IDE', 'DE', 'E']

        >>> build_fragment_sequences("PEPTIDE", "o")
        Traceback (most recent call last):
        ValueError: Ion type o is invalid.

    """

    return list(sequence_trimmer(sequence, forward=is_forward(ion_type)))


def create_fragment_internal_sequences(sequence: str, ion_type: str) -> List[str]:
    """
   Builds all internal fragment subsequences for the input `sequence`.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param ion_type: The type of ion for which internal fragments are generated.
    :type ion_type: str

    :raise ValueError: If the ion type is not one of 'a', 'b', 'c', 'x', 'y', or 'z'.

    :return: The internal fragment subsequences.
    :rtype: List[str]

    .. code-block:: python

        >>> create_fragment_internal_sequences("PEPTIDE", "a")
        ['PEPTIDE', 'EPTIDE', 'PTIDE', 'TIDE', 'IDE', 'DE', 'E']

        >>> create_fragment_internal_sequences("PEPTIDE", "y")
        ['PEPTIDE', 'PEPTID', 'PEPTI', 'PEPT', 'PEP', 'PE', 'P']

        >>> create_fragment_internal_sequences("PEPTIDE", "o")
        Traceback (most recent call last):
            ...
        ValueError: Ion type o is invalid.

    """

    return list(sequence_trimmer(sequence, forward=not is_forward(ion_type)))


def sequence_trimmer(sequence: str, forward: bool) -> Generator[str, None, None]:
    """
    A generator that yields all subsequences of the input `sequence`, by trimming from the front or back.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param forward: If True, generate sequences from the front. If False, generate sequences from the back.
    :type forward: bool

    :return: A generator of sequences.
    :rtype: Generator[str, None, None]

    .. code-block:: python

        >>> list(sequence_trimmer("PEPTIDE", True))
        ['PEPTIDE', 'EPTIDE', 'PTIDE', 'TIDE', 'IDE', 'DE', 'E']

        >>> list(sequence_trimmer("PEPTIDE", False))
        ['PEPTIDE', 'PEPTID', 'PEPTI', 'PEPT', 'PEP', 'PE', 'P']

        # Works with modifications
        >>> list(sequence_trimmer("[Acetyl]PE(3.0)PTIDE", True))
        ['[Acetyl]PE(3.0)PTIDE', 'E(3.0)PTIDE', 'PTIDE', 'TIDE', 'IDE', 'DE', 'E']

        >>> list(sequence_trimmer("PEP(1.0)TIDE[Amide]", False))
        ['PEP(1.0)TIDE[Amide]', 'PEP(1.0)TID', 'PEP(1.0)TI', 'PEP(1.0)T', 'PEP(1.0)', 'PE', 'P']

    """

    if forward is True:
        return _trim_from_start(sequence)

    return _trim_from_end(sequence)


def _trim_from_end(sequence: str) -> Generator[str, None, None]:
    """
    This function yields sub-sequences progressively by removing the last amino acid or modification
    from the given sequence until the sequence is empty. If the last character in the sequence is a
    closing parenthesis (indicating a modification), the entire modification is removed. This function
    generates the sequences associated with abc ions.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str

    :yield: The sequence with the back amino acid (and modification) removed.
    :rtype: Generator[str, None, None]

    .. code-block:: python

        >>> list(_trim_from_end('PET'))
        ['PET', 'PE', 'P']
        
        >>> list(_trim_from_end('[Acetyl]P(3.14)ET[Amide]'))
        ['[Acetyl]P(3.14)ET[Amide]', '[Acetyl]P(3.14)E', '[Acetyl]P(3.14)']

    """

    while sequence:
        yield sequence
        sequence = strip_c_term(sequence)


def _trim_from_start(sequence: str) -> Generator[str, None, None]:
    """
    This function yields sub-sequences progressively by removing the first amino acid or modification
    from the given sequence until the sequence is empty. If the first character(s) in the sequence
    indicate a modification (enclosed in parentheses), the entire modification is removed. This function
    generates the sequences associated with xyz ions.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str

    :yield: The sequence with the front amino acid (and modification) removed.
    :rtype: Generator[str, None, None]

    .. code-block:: python

        >>> list(_trim_from_start('PET'))
        ['PET', 'ET', 'T']

        >>> list(_trim_from_start('[Acetyl]P(3.14)ET[Amide]'))
        ['[Acetyl]P(3.14)ET[Amide]', 'ET[Amide]', 'T[Amide]']

    """

    while sequence:
        yield sequence
        sequence = strip_n_term(sequence)


def fragment(sequence: str, ion_types: Union[str, List[str]] = 'y', charges: Union[int, List[int]] = 1,
             monoisotopic: bool = True, internal: bool = False) -> Generator[float, None, None]:
    """
    Generates fragment ion mz values.

    :param sequence: The amino acid sequence, which can include modifications.
    :type sequence: str
    :param ion_types: A list of ion types to consider, e.g., ['b', 'y'], or a single ion type, e.g., 'b'.
    :type ion_types: Union[List[str], str]
    :param charges: A list of charge states for the fragments, or a single charge state.
    :type charges: Union[List[int], int]
    :param monoisotopic: Indicates whether to calculate monoisotopic mass, defaults to [True].
    :type monoisotopic: bool
    :param internal: Indicates whether to calculate internal fragments, defaults to [False].
    :type internal: bool

    :raise ValueError: If the ion type is not one of 'a', 'b', 'c', 'x', 'y', or 'z'.

    :yield: The calculated m/z for the fragment.
    :rtype: float

    .. code-block:: python

        >>> list(fragment("TIDE", ion_types="y", charges=1))
        [477.21911970781, 376.17144123940005, 263.08737726227, 148.06043423844]

        >>> list(fragment("TIDE", ion_types="b", charges=2))
        [230.10791574544, 165.58661920145502, 108.07314768954, 51.531115700975]

        >>> list(fragment("[1.0]T(-1.0)IDE(2.0)[-2.0]", ion_types="y", charges=1))
        [477.21911970781, 376.17144123940005, 263.08737726227, 148.06043423844]

        >>> list(fragment("[Acetyl]TIDE", ion_types="b", charges=2))
        Traceback (most recent call last):
        ...
        ValueError: could not convert string to float: 'Acetyl'

        # Can generate internal fragments too, but they are not very useful.
        # Here is the order of fragments generated:
        # [PEP, PE, P,  PEP,  E,  P]
        # [y3, y3i, y3i, y2, y2i, y1]
        >>> list(fragment("PEP", ion_types="y", charges=1, internal=True))
        [342.16596193614004, 245.11319808729002, 116.07060499932, 245.11319808729002, 148.06043423844, 116.07060499932]

    """

    if isinstance(ion_types, str):
        ion_types = [ion_types]

    if isinstance(charges, int):
        charges = [charges]

    def _fragment(t: str, c: int):
        # Loop through the sequence to generate fragments and yield m/z
        for sub_sequence in sequence_trimmer(sequence, forward=is_forward(t)):
            yield calculate_mz(sequence=sub_sequence, charge=c, ion_type=t, monoisotopic=monoisotopic)
            if internal is True:
                for i, internal_sub_sequence in enumerate(
                        create_fragment_internal_sequences(sequence=sub_sequence, ion_type=t)):
                    if i == 0:  # skip first since its already created
                        continue
                    yield calculate_mz(sequence=internal_sub_sequence, charge=c, ion_type=t,
                                       monoisotopic=monoisotopic)

    for ion_type in ion_types:
        for charge in charges:
            yield from _fragment(ion_type, charge)



"""
Sequence.py - Amino Acid Sequence Manipulation with Modifications

This module provides a set of utility functions to work with amino acid sequences that may contain modifications.
Such modifications can occur at any position within the sequence and can also be present at the N-terminus or
C-terminus of a peptide sequence.

Key Features:
- Parsing sequences with modifications and extracting modification details.
- Calculating the length of sequences excluding any modification notations.
- Adding, shifting, reversing, and stripping modifications from sequences.
- Applying both static and variable modifications to sequences.
- Validating the structure and integrity of modified sequences.

Modification Notations:
- Term modifications (N-terminus and C-terminus) are specified using square brackets: `[]`.
- Residue modifications within the sequence are specified using parentheses: `()`.

Modification Types and Representation:
- Modifications can be represented as strings (e.g., "Phospho", "Acetyl"), integers, or floats.
- String-based modifications typically represent the name or type of the modification.
- Numerical modifications (integers or floats) represent specific mass changes associated with a modification.
- During parsing, the module automatically identifies the modification type based on its representation.

Example Sequences:
    [Acetyl]P(Oxydation)EPTI(Phospho)DE[Amide] or [1.234]PE(1.0)PTI(2.0)DE[3.1415]

The module ensures that sequences and their modifications are handled correctly, preserving the positional
relationship between amino acids and their modifications during operations.
"""

import re
from copy import deepcopy
from typing import Dict, List, Any, Generator, Union, Set, Tuple

from peptacular.term import get_c_term_modification, strip_term_modifications, add_n_term_modification, \
    add_c_term_modification, get_n_term_modification
from peptacular.util import convert_type, identify_regex_indexes, _validate_span


def calculate_sequence_length(sequence: str) -> int:
    """
    Compute the length of the peptide sequence, excluding any modification notation.

    :param sequence: The amino acid sequence
    :type sequence: str
    :return: Length of the unmodified sequence.
    :rtype: int

    .. code-block:: python

        >>> calculate_sequence_length("[Acetyl]PEP(1.2345)TID(3.14)E[Amide]")
        7
        >>> calculate_sequence_length("PEPTIDE")
        7

    """

    return len(strip_modifications(sequence))


def get_modifications(sequence: str) -> Dict[int, Union[str, float, int]]:
    """
    Parses a peptide sequence with modifications and returns a dictionary where keys
    represent the position of the modified amino acid and values are the respective modifications.

    :param sequence: The peptide sequence, potentially including modifications.
    :type sequence: str
    :return: A dictionary with the modification values indexed by the position
             of the modified amino acid.
    :rtype: Dict[int, Union[str, int, float]]

    .. code-block:: python

        # All modifications will be returned as either strings, ints or floats.
        >>> get_modifications('PEP(Phospho)T(1)IDE(3.14)')
        {2: 'Phospho', 3: 1, 6: 3.14}

        # N-terminus and C-terminal modifications are index by -1, and the length of the sequence, respectively:
        >>> get_modifications('[Acetyl]PEPTIDE[Amide]')
        {-1: 'Acetyl', 7: 'Amide'}

        # A sequecnes without any modifications will return an empty dictionary:
        >>> get_modifications('PEPTIDE')
        {}

    """

    n_term_mod = get_n_term_modification(sequence)
    c_term_mod = get_c_term_modification(sequence)

    sequence = strip_term_modifications(sequence)
    stripped_sequence = strip_modifications(sequence)

    mod_pattern = re.compile(r'\(([^)]*)\)')
    matches = mod_pattern.finditer(sequence)

    modifications = {}

    if n_term_mod:
        modifications[-1] = n_term_mod

    mod_offset = 0

    for match in matches:
        mod_index = match.start() - mod_offset - 1
        mod = convert_type(match.group(1))
        modifications[mod_index] = mod

        # Compute offset for the next modification
        mod_offset += len(match.group())

    if c_term_mod:
        modifications[len(stripped_sequence)] = c_term_mod

    return modifications


def _construct_sequence_with_modifications(sequence: str, mod_map: Dict[int, Any]) -> str:
    """
    Builds a modified peptide sequence from an unmodified sequence and a dictionary of modifications.
    """

    n_term_mod = mod_map.pop(-1, None)
    c_term_mod = mod_map.pop(calculate_sequence_length(sequence), None)

    modified_sequence = []
    prev_index = 0

    # Sort the modifications by index in descending order
    for mod_index, mod in sorted(mod_map.items()):

        if mod_index < -1 or mod_index >= len(sequence):
            raise ValueError(f'Index of modification: {mod_index} is invalid for peptide sequence: {sequence}')

        # Insert the modification into the modified peptide sequence
        modified_sequence.append(sequence[prev_index: mod_index + 1])
        modified_sequence.append(f"({mod})")
        prev_index = mod_index + 1

    modified_sequence.append(sequence[prev_index:])
    modified_sequence = ''.join(modified_sequence)
    modified_sequence = add_n_term_modification(modified_sequence, n_term_mod)
    modified_sequence = add_c_term_modification(modified_sequence, c_term_mod)
    return modified_sequence


def add_modifications(sequence: str, modifications: Dict[int, Any], overwrite: bool = True) -> str:
    """
    Adds modifications to the given peptide sequence.

    :param sequence: Unmodified peptide sequence.
    :type sequence: str
    :param modifications: Dictionary with indices of the modified amino acids and corresponding modifications.
    :type modifications: Dict[int, Any]
    :param overwrite: If True, existing modifications will be overwritten. If False, existing modifications will be
                        preserved.
    :type overwrite: bool

    :return: The peptide sequence with the specified modifications.
    :rtype: str

    .. code-block:: python

        # Add internal modifications to an unmodified peptide
        >>> add_modifications('PEPTIDE', {2: 'phospho'})
        'PEP(phospho)TIDE'

        # Can also add N and C terminal modifications
        >>> add_modifications('PEPTIDE', {-1: 'Acetyl', 6: '1.234', 7: 'Amide'})
        '[Acetyl]PEPTIDE(1.234)[Amide]'

        # Empty modification dicts will simply return the input sequence
        >>> add_modifications('PEPTIDE', {})
        'PEPTIDE'

        # Existing modifications will be overwritten by default
        >>> add_modifications('PEP(phospho)TIDE', {2: 'acetyl'})
        'PEP(acetyl)TIDE'

        # Can also preserve existing modifications
        >>> add_modifications('PEP(phospho)TIDE', {2: 'acetyl'}, overwrite=False)
        'PEP(phospho)TIDE'

    """

    stripped_sequence = strip_modifications(sequence)
    original_mods = get_modifications(sequence)

    for mod_index, mod in sorted(modifications.items()):
        if mod_index in original_mods:
            if overwrite is True:
                original_mods[mod_index] = mod
        else:
            original_mods[mod_index] = mod

    return _construct_sequence_with_modifications(stripped_sequence, original_mods)


def strip_modifications(sequence: str) -> str:
    """
    Removes modifications from the given sequence.

    :param sequence: modified sequence to be stripped
    :type sequence: str

    :return: unmodified sequence
    :rtype: str

    .. code-block:: python

        # Removes internal modifications:
        >>> strip_modifications('PEP(phospho)TIDE')
        'PEPTIDE'

        # Also removes N and C terminal modifications:
        >>> strip_modifications('[Acetyl](3.14)PEPTIDE(1.234)[Amide]')
        'PEPTIDE'

        # Using a sequence without modifications will return the same sequence:
        >>> strip_modifications('PEPTIDE')
        'PEPTIDE'

    """

    # Removing modifications in parentheses
    sequence = re.sub(r'\([^)]*\)', '', sequence)

    # Removing modifications in square brackets
    sequence = re.sub(r'\[[^]]*]', '', sequence)

    return sequence


def apply_static_modifications(sequence: str, mod_map: Dict[str, Any], overwrite: bool = True) -> str:
    """
    Add static modifications to an amino acid sequence. If a modification is already present in the sequence,
    it will be replaced by the new one.

    :param sequence: Original amino acid sequence.
    :type sequence: str
    :param mod_map: Dictionary mapping amino acids to the mass of their modifications.
    :type mod_map: Dict[str, Any]
    :param overwrite: If True, existing modifications will be replaced by the new ones.
    :type overwrite: bool

    :return: Modified amino acid sequence.
    :rtype: str

    .. code-block:: python

        # Applies static modifcation to all matching residues in the sequence:
        >>> apply_static_modifications('PEPTIDE', {'P': 'phospho'})
        'P(phospho)EP(phospho)TIDE'

        # Can specify multiple static modifications:
        >>> apply_static_modifications('PEPTIDE', {'P': 'phospho', 'E': '3.0'})
        'P(phospho)E(3.0)P(phospho)TIDE(3.0)'

        # Works with already modified sequences:
        >>> apply_static_modifications('[3.14]PEPTIDE(1.234)', {'P': 'phospho'})
        '[3.14]P(phospho)EP(phospho)TIDE(1.234)'

        # By default, any overlapping modifications will be replaced:
        >>> apply_static_modifications('PEP(1.234)TIDE', {'P': 'phospho'})
        'P(phospho)EP(phospho)TIDE'

        # To not reaplce existing modifications, set override to False:
        >>> apply_static_modifications('PEP(1.234)TIDE', {'P': 'phospho'}, overwrite=False)
        'P(phospho)EP(1.234)TIDE'

        # Can also use regular expressions to match multiple residues:
        >>> apply_static_modifications('PEPTIDE', {'(?<=P)E': 'phospho'})
        'PE(phospho)PTIDE'

        >>> apply_static_modifications('PEPTIDE', {'PE': 'phospho'})
        'P(phospho)EPTIDE'

    """

    stripped_sequence = strip_modifications(sequence)
    original_mod_map = get_modifications(sequence)

    for regex_str, mod_mass in mod_map.items():
        for mod_index in identify_regex_indexes(stripped_sequence, regex_str):
            if overwrite is True or mod_index not in original_mod_map:
                original_mod_map[mod_index] = mod_mass

    return add_modifications(stripped_sequence, original_mod_map)


def _apply_variable_modifications_rec(mods: Dict[int, Set], sequence: str, index: int, current_mods: Dict[int, Any],
                                      max_mod_count: int) -> Generator[Dict[int, Any], None, None]:
    """
    Apply variable modifications to a sequence, generating all possible combinations.

    :param mods: Dictionary mapping amino acids to the mass of their modifications.
    :type mods: Dict[int, Any]
    :param sequence: the unmodified sequence.
    :type sequence: str
    :param index: Current index on the amino acid sequence.
    :type index: int
    :param current_mods: Dictionary mapping the indices of the amino acids to the mass of their modifications.
    :type current_mods: Dict[int, str]
    :param max_mod_count: Maximum number of modifications allowed on the amino acid sequence.
    :type max_mod_count: int

    :yield: Modified dictionary mapping the indices of the amino acid sequence to the mass of their modifications.
    :rtype: Generator[Dict[int, float], None, None]
    """

    if index == len(sequence) or len(current_mods) == max_mod_count:
        yield current_mods
        return

    original_mod_dict = deepcopy(current_mods)

    curr_mods = mods.get(index, None)

    if curr_mods is not None:
        for curr_mod in curr_mods:
            updated_mod_dict = deepcopy(current_mods)
            if index not in current_mods:
                updated_mod_dict[index] = curr_mod
                yield from _apply_variable_modifications_rec(mods, sequence, index + 1, updated_mod_dict,
                                                             max_mod_count)
    yield from _apply_variable_modifications_rec(mods, sequence, index + 1, original_mod_dict, max_mod_count)


def apply_variable_modifications(sequence: str, mod_map: Dict[str, Any], max_mods: int) -> List[str]:
    """
    Apply variable modifications to a sequence.

    :param sequence: Original amino acid sequence.
    :type sequence: str
    :param mod_map: Dictionary mapping amino acids to the mass of their modifications.
    :type mod_map: Dict[str, Any]
    :param max_mods: Maximum number of modifications allowed on the peptide.
    :type max_mods: int

    :return: List of all possible modified peptide sequences.
    :rtype: List[str]

    .. code-block:: python

        >>> apply_variable_modifications('PEPTIDE', {'P': 'phospho'}, 2)
        ['P(phospho)EP(phospho)TIDE', 'P(phospho)EPTIDE', 'PEP(phospho)TIDE', 'PEPTIDE']

        # Setting max_mods to 0 will return the original sequence:
        >>> apply_variable_modifications('PEPTIDE', {'P': 'phospho'}, 0)
        ['PEPTIDE']

        # Works with already modified sequences, and will not replace existing modifications:
        >>> apply_variable_modifications('[Acetyl]P(3.14)EPTIDE[Amide]', {'P': 'phospho'}, 2)
        ['[Acetyl]P(3.14)EP(phospho)TIDE[Amide]', '[Acetyl]P(3.14)EPTIDE[Amide]']

        # Can also use regular expressions to match multiple residues:
        >>> apply_variable_modifications('PEPTIDE', {'(?<=P)E': 'phospho'}, 2)
        ['PE(phospho)PTIDE', 'PEPTIDE']

    """

    original_mods = get_modifications(sequence)
    stripped_sequence = strip_modifications(sequence)

    new_mod_map: Dict[int, Set] = {}
    for regex_str, mod_mass in mod_map.items():
        for mod_index in identify_regex_indexes(stripped_sequence, regex_str):
            new_mod_map.setdefault(mod_index, set()).add(mod_mass)

    mods = _apply_variable_modifications_rec(new_mod_map, stripped_sequence, 0, original_mods,
                                             max_mods + len(original_mods))
    return [add_modifications(stripped_sequence, mod) for mod in mods]


def reverse_sequence(sequence: str, swap_terms: bool = False) -> str:
    """
    Reverses the sequence, while preserving the position of any modifications.

    :param sequence: The amino acid sequence to be reversed.
    :type sequence: str
    :param swap_terms: If True, the N- and C-terminal modifications will be swapped.
    :type swap_terms: bool

    :return: The reversed sequence with modifications preserved.
    :rtype: str

    .. code-block:: python

        # For unmodified sequences, the result is the same as the built-in string reversal:
        >>> reverse_sequence('PEPTIDE')
        'EDITPEP'

        # For modified sequences, the modifications are preserved on the associated residues:
        >>> reverse_sequence('[Acetyl]P(phospho)EP(phospho)TIDE[Amide]')
        '[Acetyl]EDITP(phospho)EP(phospho)[Amide]'

        # If swap_terms is True, the N- and C-terminal modifications will be swapped too:
        >>> reverse_sequence('[Acetyl]P(phospho)EP(phospho)TIDE[Amide]', swap_terms=True)
        '[Amide]EDITP(phospho)EP(phospho)[Acetyl]'

    """

    mods = get_modifications(sequence)
    stripped_sequence = strip_modifications(sequence)[::-1]

    n_term = mods.pop(-1, None)
    c_term = mods.pop(len(stripped_sequence), None)

    mod_reverse = {len(stripped_sequence) - 1 - k: v for k, v in mods.items()}

    modified_sequence = add_modifications(stripped_sequence, mod_reverse)

    if swap_terms:
        modified_sequence = add_n_term_modification(modified_sequence, c_term)
        modified_sequence = add_c_term_modification(modified_sequence, n_term)

    else:
        modified_sequence = add_n_term_modification(modified_sequence, n_term)
        modified_sequence = add_c_term_modification(modified_sequence, c_term)

    return modified_sequence


def shift_sequence(sequence: str, shift: int) -> str:
    """
    Shifts the sequence to the left by a given number of positions, while preserving the position of any modifications.

    :param sequence: The sequence to be shifted.
    :type sequence: str
    :param shift: The number of positions to shift the sequence to the left.
    :type shift: int

    :return: The shifted sequence with modifications preserved.
    :rtype: str

    .. code-block:: python

        >>> shift_sequence('PEPTIDE', 2)
        'PTIDEPE'

        >>> shift_sequence('[Acetyl]P(phospho)EP(phospho)TIDE[Amide]', 2)
        '[Acetyl]P(phospho)TIDEP(phospho)E[Amide]'

        # Shifting by 0 or length of sequence positions returns the original sequence:
        >>> shift_sequence('[Acetyl]P(phospho)EP(phospho)TIDE[Amide]', 7)
        '[Acetyl]P(phospho)EP(phospho)TIDE[Amide]'
        >>> shift_sequence('[Acetyl]P(phospho)EP(phospho)TIDE[Amide]', 0)
        '[Acetyl]P(phospho)EP(phospho)TIDE[Amide]'

        # Shifting by a negative number shifts the sequence to the right:
        >>> shift_sequence('[Acetyl]P(phospho)EP(phospho)TIDE[Amide]', -2)
        '[Acetyl]DEP(phospho)EP(phospho)TI[Amide]'

    """

    mods = get_modifications(sequence)
    stripped_sequence = strip_modifications(sequence)

    n_term = mods.pop(-1, None)
    c_term = mods.pop(len(stripped_sequence), None)

    seq_len = len(stripped_sequence)

    # Take modulus to ensure proper wrapping around
    effective_shift = shift % seq_len

    shifted_stripped_sequence = stripped_sequence[effective_shift:] + stripped_sequence[:effective_shift]

    # Shift mod positions, taking care to avoid negative keys
    shifted_sequence_mods = {(k - effective_shift) % seq_len if k != -1 else -1: v for k, v in mods.items()}

    modified_sequence = add_modifications(shifted_stripped_sequence, shifted_sequence_mods)
    modified_sequence = add_n_term_modification(modified_sequence, n_term)
    modified_sequence = add_c_term_modification(modified_sequence, c_term)

    return modified_sequence


def is_sequence_valid(sequence: str) -> bool:
    """
    Checks if the sequence is valid. A sequence is valid if it does not contain doubly enclosed modifications.

    :param sequence: modified sequence
    :type sequence: str

    :return: True, if the sequence is valid, False otherwise
    :rtype: bool

    .. code-block:: python

        >>> is_sequence_valid('PEPTIDE')
        True

        >>> is_sequence_valid('[Acetyl]P(phospho)EP(phospho)TIDE[Amide]')
        True

        >>> is_sequence_valid('[Acetyl]P(phospho)EP(phospho)TIDE[Amide]]')
        False

        >>> is_sequence_valid('[Acetyl]P(phospho)EP((phospho))TIDE[Amide]')
        False

    """

    stripped_sequence = strip_modifications(sequence)
    residues = set(stripped_sequence)

    # check if parenthesis are in the stripped sequence
    if '(' in residues or ')' in residues:
        return False

    # check if square brackets are in the stripped sequence
    if '[' in residues or ']' in residues:
        return False

    # check modifications
    mods = get_modifications(sequence)
    for _, mod in mods.items():
        if isinstance(mod, str):
            if '(' in mod or ')' in mod:
                return False
            if '[' in mod or ']' in mod:
                return False

    return True


def span_to_sequence(sequence: str, span: Tuple[int, int, int]) -> str:
    """
    Extracts a subsequence from the input sequence based on the provided span.

    :param sequence: The original sequence.
    :type sequence: str
    :param span: A tuple representing the span of the subsequence to be extracted.
    :type span: Tuple[int, int, int]

    :return: The subsequence of the input sequence defined by the span.
    :rtype: str

    .. code-block:: python

        # Works with unmodified sequences
        >>> span_to_sequence('PEPTIDE', (0, 4, 0))
        'PEPT'

        >>> span_to_sequence('PEPTIDE', (1, 6, 0))
        'EPTID'

        >>> span_to_sequence('PEPTIDE', (4, 7, 0))
        'IDE'

        # but will also preserve modifications, including terminal modifications
        >>> span_to_sequence('[Acetyl]P(1.0)EPTIDE', (0, 4, 0))
        '[Acetyl]P(1.0)EPT'

        >>> span_to_sequence('PEPTIDE(1.0)[Amide]', (4, 7, 0))
        'IDE(1.0)[Amide]'

    """

    _validate_span(span)

    stripped_sequence = strip_modifications(sequence)

    base_sequence = stripped_sequence[span[0]:span[1]]

    # if peptide is unmodified skip the modification parsing
    if stripped_sequence == sequence:
        return base_sequence

    mods = get_modifications(sequence)
    n_term = get_n_term_modification(sequence)
    c_term = get_c_term_modification(sequence)

    mods = {k-span[0]: mods[k] for k in mods if span[0] <= k < span[1]}

    if span[0] == 0 and n_term is not None:
        mods[-1] = n_term

    if span[1] == len(stripped_sequence) and c_term is not None:
        mods[len(base_sequence)] = c_term

    return add_modifications(base_sequence, mods)



import unittest

from peptacular import constants
from peptacular.digest import digest, identify_cleavage_sites, build_right_semi_sequences, \
    build_non_enzymatic_sequences, build_left_semi_sequences, build_semi_sequences

PROTEIN = 'MVIMSEFSADPAGQGQGQQKPLRVGFYDIERTLGKGNFAVVKLARHRVTKTQVAIKIIDKTRLDSSNLEKIYREVQLMKLLNHPHIIKLYQVMETKDMLYIVTE'


class TestDigest(unittest.TestCase):

    def test_trypsin_sites(self):
        cleavage_sites = identify_cleavage_sites(PROTEIN, constants.PROTEASES['trypsin'])

        sites = [23, 31, 35, 42, 45, 47, 50, 56, 60, 62, 70, 73, 79, 88, 96]

        self.assertEqual(cleavage_sites, sites)

    def test_digest_protein(self):
        peptides = set(digest(sequence='TIDERTIDEKTIDE',
                              enzyme_regex=constants.PROTEASES['trypsin'],
                              missed_cleavages=2,
                              min_len=0,
                              max_len=100,
                              semi=False))

        self.assertEqual(peptides,
                         {'TIDER', 'TIDERTIDEK', 'TIDERTIDEKTIDE', 'TIDEK', 'TIDEKTIDE', 'TIDE'})

        peptides = set(digest(sequence='TIDERTIDEKTIDE',
                              enzyme_regex=constants.PROTEASES['trypsin'],
                              missed_cleavages=1,
                              min_len=0,
                              max_len=100,
                              semi=False))
        self.assertEqual(peptides, {'TIDER', 'TIDERTIDEK', 'TIDEK', 'TIDEKTIDE', 'TIDE'})

        peptides = set(digest(sequence='KTIDERTIDEKTIDE',
                              enzyme_regex=constants.PROTEASES['trypsin'],
                              missed_cleavages=1,
                              min_len=0,
                              max_len=100,
                              semi=False))
        self.assertEqual(peptides,
                         {'K', 'KTIDER', 'TIDER', 'TIDERTIDEK', 'TIDEK', 'TIDEKTIDE', 'TIDE'})

        peptides = set(digest(sequence='TIDERTIDEKTIDEK',
                              enzyme_regex=constants.PROTEASES['trypsin'],
                              missed_cleavages=1,
                              min_len=0,
                              max_len=100,
                              semi=False))
        self.assertEqual(peptides, {'TIDER', 'TIDERTIDEK', 'TIDEK', 'TIDEKTIDEK', 'TIDEK'})

        peptides = set(digest(sequence='TIDERTIDEKKTIDE',
                              enzyme_regex=constants.PROTEASES['trypsin'],
                              missed_cleavages=1,
                              min_len=0,
                              max_len=100,
                              semi=False))
        self.assertEqual(peptides,
                         {'TIDER', 'TIDERTIDEK', 'TIDEK', 'TIDEKK', 'K', 'KTIDE', 'TIDE'})

        peptides = set(digest(sequence='TIDERTIDEKTIDE',
                              enzyme_regex=constants.PROTEASES['trypsin'],
                              missed_cleavages=0,
                              min_len=0,
                              max_len=100,
                              semi=False))
        self.assertEqual(peptides, {'TIDER', 'TIDEK', 'TIDE'})

        peptides = set(digest(sequence='TIDERTIDEKTIDE',
                              enzyme_regex=constants.PROTEASES['trypsin'],
                              missed_cleavages=10,
                              min_len=0,
                              max_len=100,
                              semi=False))
        self.assertEqual(peptides,
                         {'TIDER', 'TIDERTIDEK', 'TIDERTIDEKTIDE', 'TIDEK', 'TIDEKTIDE', 'TIDE'})

    def test_get_semi_sequences(self):
        assert set(build_semi_sequences(sequence='PEPTIDE', min_len=3, max_len=5)) == \
               {'TIDE', 'PTIDE', 'PEPT', 'PEPTI', 'IDE', 'PEP'}
        assert set(build_semi_sequences(sequence='PEPTIDE', min_len=2, max_len=5)) == \
               {'TIDE', 'PTIDE', 'PEPT', 'PEPTI', 'IDE', 'PEP', 'PE', 'DE'}
        assert set(build_semi_sequences(sequence='PEPTIDE', min_len=2, max_len=100)) == \
               {'TIDE', 'PTIDE', 'PEPT', 'PEPTI', 'IDE', 'PEP', 'PE', 'DE', 'PEPTID', 'EPTIDE'}
        assert set(build_semi_sequences(sequence='PEPTIDE', min_len=None, max_len=None)) == \
               {'P', 'E', 'TIDE', 'PTIDE', 'PEPT', 'PEPTI', 'IDE', 'PEP', 'PE', 'DE', 'PEPTID', 'EPTIDE'}

    def test_get_left_semi_sequences(self):
        self.assertEqual(set(build_left_semi_sequences('PEPTIDE', None, None)),
                         {'P', 'PE', 'PEP', 'PEPT', 'PEPTI', 'PEPTID'})
        self.assertEqual(set(build_left_semi_sequences('PEPTIDE', 3, None)), {'PEP', 'PEPT', 'PEPTI', 'PEPTID'})
        self.assertEqual(set(build_left_semi_sequences('PEPTIDE', None, 5)), {'P', 'PE', 'PEP', 'PEPT', 'PEPTI'})

        assert set(build_left_semi_sequences(sequence='PEPTIDE', min_len=None, max_len=None)) == \
               {'P', 'PE', 'PEP', 'PEPT', 'PEPTI', 'PEPTID'}
        assert set(build_left_semi_sequences(sequence='PEPTIDE', min_len=3, max_len=None)) == \
               {'PEP', 'PEPT', 'PEPTI', 'PEPTID'}
        assert set(build_left_semi_sequences(sequence='PEPTIDE', min_len=None, max_len=5)) == \
               {'P', 'PE', 'PEP', 'PEPT', 'PEPTI'}
        assert set(build_left_semi_sequences(sequence='PEPTIDE', min_len=3, max_len=4)) == \
               {'PEP', 'PEPT'}

    def test_get_right_semi_sequences(self):
        self.assertEqual(set(build_right_semi_sequences('PEPTIDE', None, None)),
                         {'EPTIDE', 'PTIDE', 'TIDE', 'IDE', 'DE', 'E'})
        self.assertEqual(set(build_right_semi_sequences('PEPTIDE', 3, None)), {'EPTIDE', 'PTIDE', 'TIDE', 'IDE'})
        self.assertEqual(set(build_right_semi_sequences('PEPTIDE', None, 5)), {'PTIDE', 'TIDE', 'IDE', 'DE', 'E'})
        assert set(build_right_semi_sequences(sequence='PEPTIDE', min_len=None, max_len=None)) == \
               {'EPTIDE', 'PTIDE', 'TIDE', 'IDE', 'DE', 'E'}
        assert set(build_right_semi_sequences(sequence='PEPTIDE', min_len=3, max_len=None)) == \
               {'EPTIDE', 'PTIDE', 'TIDE', 'IDE'}
        assert set(build_right_semi_sequences(sequence='PEPTIDE', min_len=None, max_len=5)) == \
               {'PTIDE', 'TIDE', 'IDE', 'DE', 'E'}
        assert set(build_right_semi_sequences(sequence='PEPTIDE', min_len=3, max_len=4)) == \
               {'TIDE', 'IDE'}

    def test_get_all_non_enzymatic_peptides(self):
        self.assertEqual(set(build_non_enzymatic_sequences('PEPT')),
                         {'P', 'E', 'P', 'T', 'PE', 'EP', 'PT', 'PEP', 'EPT'})
        self.assertEqual(set(build_non_enzymatic_sequences('PEPT', min_len=1, max_len=2)),
                         {'P', 'E', 'P', 'T', 'PE', 'EP', 'PT'})
        self.assertEqual(set(build_non_enzymatic_sequences('PEPT', min_len=2, max_len=4)),
                         {'PE', 'EP', 'PT', 'PEP', 'EPT'})



from peptacular.protein import build_coverage_array, find_peptide_indexes, \
    calculate_percent_coverage

import unittest

class TestProtein(unittest.TestCase):

    def test_get_peptide_indexes_in_protein(self):
        # Test basic functionality
        self.assertEqual(find_peptide_indexes("AAPEPTIDEAA", "PEPTIDE"), [2])
        # Test with multiple occurrences
        self.assertEqual(find_peptide_indexes("AAPEPTIDEPEPTIDEAA", "PEPTIDE"), [2, 9])
        # Test peptide not in protein
        self.assertEqual(find_peptide_indexes("AAPEPTIDEAA", "XYZ"), [])

    def test_calculate_protein_coverage(self):
        # Test basic functionality
        protein = "AAPEPTIDEAA"
        peptides = ["PEPTIDE"]
        self.assertEqual(build_coverage_array(protein, peptides), [0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0])

        # Test with multiple peptides
        peptides = ["PEP", "TIDE"]
        self.assertEqual(build_coverage_array(protein, peptides), [0, 0, 1, 1, 1, 1, 1, 1, 1, 0, 0])

        # Test with peptide not in protein
        peptides = ["XYZ"]
        self.assertEqual(build_coverage_array(protein, peptides), [0] * 11)

    def test_calculate_protein_coverage_percent(self):
        # Test basic functionality
        protein = "AAPEPTIDEAA"
        peptides = ["PEPTIDE"]
        self.assertAlmostEqual(calculate_percent_coverage(protein, peptides), 7 / 11)

        # Test with multiple peptides
        peptides = ["PEP", "TIDE"]
        self.assertAlmostEqual(calculate_percent_coverage(protein, peptides), 7 / 11)

        # Test with peptide not in protein
        peptides = ["XYZ"]
        self.assertAlmostEqual(calculate_percent_coverage(protein, peptides), 0 / 11)

    def test_overlapping_peptides(self):
        protein = "AAPEPTIDEPEPTIDEAA"
        peptides = ["PEPTIDEPEP"]
        self.assertEqual(build_coverage_array(protein, peptides),
                         [0, 0, 1, 1, 1, 1, 1, 1, 1, 1, 1, 1, 0, 0, 0, 0, 0, 0])

    def test_empty_inputs(self):
        # Empty protein and peptides
        self.assertEqual(find_peptide_indexes("", ""), [])
        self.assertEqual(build_coverage_array("", []), [])
        self.assertEqual(calculate_percent_coverage("", []), 0)

        # Empty protein
        self.assertEqual(find_peptide_indexes("", "PEP"), [])
        self.assertEqual(build_coverage_array("", ["PEP"]), [])
        self.assertEqual(calculate_percent_coverage("", ["PEP"]), 0)

        # Empty peptide
        self.assertEqual(find_peptide_indexes("AAPEPTIDEAA", ""), [])
        self.assertEqual(build_coverage_array("AAPEPTIDEAA", [""]), [0] * 11)
        self.assertEqual(calculate_percent_coverage("AAPEPTIDEAA", [""]), 0)

    def test_peptide_longer_than_protein(self):
        # Peptide is longer than the protein
        self.assertEqual(find_peptide_indexes("PEP", "PEPTIDE"), [])
        self.assertEqual(build_coverage_array("PEP", ["PEPTIDE"]), [0, 0, 0])
        self.assertEqual(calculate_percent_coverage("PEP", ["PEPTIDE"]), 0)

    def test_full_coverage(self):
        # Peptides fully cover the protein
        peptides = ["AAPEP", "TIDEAA"]
        self.assertEqual(build_coverage_array("AAPEPTIDEAA", peptides), [1] * 11)
        self.assertEqual(calculate_percent_coverage("AAPEPTIDEAA", peptides), 1)

    def test_no_coverage(self):
        # No peptide matches in the protein
        peptides = ["XYZ", "LMN"]
        self.assertEqual(build_coverage_array("AAPEPTIDEAA", peptides), [0] * 11)
        self.assertEqual(calculate_percent_coverage("AAPEPTIDEAA", peptides), 0)

    def test_case_sensitivity(self):
        # Check if the functions are case-sensitive
        self.assertEqual(find_peptide_indexes("AAPEPTIDEAA", "peptide"), [])
        peptides = ["pep", "TIDE"]
        self.assertNotEqual(build_coverage_array("AAPEPTIDEAA", peptides), [1] * 11)
        self.assertNotEqual(calculate_percent_coverage("AAPEPTIDEAA", peptides), 1)


if __name__ == "__main__":
    unittest.main()
